# coding = utf-8

""" This file evaluates how a table detection algorithm performs. """

from math import *
from Common.Rect import Rect
from Common.Labels import *
from TableErrLabel import *

FAKE_LABEL_RATIO  = 0.9   # define at least 90% of fake table belong to certain label
AMP_LABEL_RATIO   = 0.9   # define at least 90% of amplified table content belong to certain label
SPLIT_MAX         = 3     # acceptable split number limit
MERGE_MAX         = 3     # acceptable merge number limit
FINAL_ACPT        = 4     # threshold used for distinguishing acceptable or not
RDC_SIG_OFFSET    = 7     # reduce sigmoid function offset
AMP_SIG_OFFSET    = 4     # amplify sigmoid function offset
    
class TableEvaluator:      
    def __init__(self, metrics, labels, gtPage, obPage):
        """ TableEvaluator initializer.             \n

        @remarks gts - ground-truth objects         \n
                 obs - obtained objects             \n
                 charno - page char amount          \n
                 pagebox - page crop box            \n
 
        @param  metrics  performance metrics dictionary         \n
        @param  labels   error categories label dictionary      \n
        @param  gtPage   ground-truth page content              \n
        @param  obPage   obtained page content                  \n
        
        """  
        self._metrics = metrics
        self._labels = labels
        self._gtPage = gtPage
        self._obPage = obPage
        self._gts = []
        self._obs = []
        self._charno = 0
        self._pagebox = []
        
    def _start(self):     
        # Get page content by label.           
        self._gts = self._gtPage.getByLabel(LABEL_TABLE)
        self._obs = self._obPage.getByLabel(LABEL_TABLE)

        for ch in LABEL_LEAF:
            self._charno += len(self._gtPage.getByLabel(ch))
        self._pagebox = self._gtPage.getCropBox()       

        
    def getScore(self):
        """ Evaluation core.         \n

        @remarks This function determines matchings between the ground-truth table objects          \n
                 and the obtained table objects according to region overlap;                        \n
                 Then six kinds of error types "fake, miss, split, merge, reduce, amplify"          \n
                 are evaluated respectively;                                                        \n
                 During this process, more detailed errors are also judged and recorded to the      \n
                 performance metrics.                                                               \n
                 Finally, a integrited penalty score is assigned as the maximum of all error scores.\n
                 
        """  
        self._start()       
        # First, judge if tables in obtained tablelist are fake detected.
        for ob in self._obs:
            mark = False
            for gt in self._gts:
                if (ob.getBox().isOverlap(gt.getBox())):
                    mark = True
                    break
            if (False == mark):
                self._evalFake(ob)
                
        # Second, determine error label for each table in groundtruth tablelist.
        if self._gts: 
            for index, gt in enumerate(self._gts):
                self._metrics['true'] += 1
                ovlap = 0
                ovlapTabs = []
                for ob in self._obs:
                    if (gt.getBox().isOverlap(ob.getBox())):
                        ovlap += 1
                        ovlapTabs.append(ob)
                if (ovlap == 0):  
                    self._evalMiss()
                else:
                    split = self._evalSplit(gt, ovlapTabs, ovlap)
                    merge = self._evalMerge(ovlapTabs)
                    reduc = self._evalReduce(gt, ovlapTabs)                   
                    amplify  = self._evalAmplify(gt, ovlapTabs)
                    final = max(split, reduc, amplify, merge)                   
                    if (final < FINAL_ACPT):
                        self._metrics['acceptable'] += 1              

 
    def _evalFake(self, ob):
        # Get all chars contained in the fake table, judge the real structure each char belong to,
        # and collect frequence of all different labels.
        obPids = self._obPage.lidToPids(ob.getLid())
        labelCount = {}
        for pid in obPids:
            gtLabel = self._mapOb2Gt(pid)          
            if gtLabel not in labelCount:
                labelCount.setdefault(gtLabel, 1)
            else:
                labelCount[gtLabel] += 1
                
        # Find out the label with largest frequence       
        items = labelCount.items()
        items.sort(key = lambda labelCount:labelCount[1], reverse = True)
        key, value = items[0]
    
        labels = {
            LABEL_LIST:     FAKE_LIST,
            LABEL_FIGURE:   FAKE_FIGURE,
            LABEL_MATRIX:   FAKE_MATRIX
        }
        strs = {
            LABEL_LIST:     'fake_list',
            LABEL_FIGURE:   'fake_figure',
            LABEL_MATRIX:   'fake_matrix'
        }

        # at lease 90% of the fake table belong to certain type in the predefined labels
        if ((value > FAKE_LABEL_RATIO * len(obPids)) & (key in labels)):
            self._labels['%s' %strs[key]] += 1
            self._metrics['fake but acceptable'] += 1
        else:
            self._labels['fake_otherObj'] += 1
            self._metrics['fake and unacceptable'] += 1
           

    def _evalMiss(self):
        self._labels['missed'] += 1
        self._metrics['missed'] += 1


    def _evalSplit(self, gt, ovlapTabs, ovlap):
        if( ovlap == 1):
            split = 0
        else:
            Rect(0,0,0,0)
            # when the table is too much splitted, its penaty score doubles
            if (Rect.distanceY(ovlapTabs[0].getBox(), ovlapTabs[1].getBox()) < 0):
                split = Error_Label[SPLIT_HOR] *  (2 if (ovlap > SPLIT_MAX) else 1)                
                self._labels['splitted_horizontal'] += 1
            else:
                split = Error_Label[SPLIT_VER] * (2 if (ovlap > SPLIT_MAX) else 1)  
                self._labels['splitted_vertical'] += 1
        return split


    def _evalMerge(self, ovlapTabs):
        mergeno = 0;
        for tab in ovlapTabs:
            ovlap = 0
            for gt in self._gts:
                if (tab.getBox().isOverlap(gt.getBox())):
                    ovlap += 1
            mergeno += ovlap - 1

        if (mergeno == 0):
            merge = 0
        else:
            merge = Error_Label[MERGE] * (2 if (mergeno > MERGE_MAX) else 1) 
            self._labels['merged'] += 1
        return merge


    def _evalReduce(self, gt, ovlapTabs):
        gtPids = self._gtPage.lidToPids(gt.getLid())
        obPids = []
        for ob in ovlapTabs:
            obPids.extend(self._obPage.lidToPids(ob.getLid()))
               
        pidRatio = float(len(frozenset(gtPids) & frozenset(obPids))) / len(gtPids)
        reduc = Error_Label[REDUCE] * (1 / float(1 + exp(10 * pidRatio - RDC_SIG_OFFSET)))
        self._labels['reduced'] += 1
        return reduc
    
         
    def _evalAmplify(self, gt, ovlapTabs):
        gtPids = self._gtPage.lidToPids(gt.getLid())
        obPids = []
        for ob in ovlapTabs:
            obPids.extend(self._obPage.lidToPids(ob.getLid())) 

        # determine gt label for each amplified pid
        labelCount = {}
        for pid in obPids:
            if pid not in gtPids:
                gtLabel = self._mapOb2Gt(pid)          
                if gtLabel not in labelCount:
                    labelCount.setdefault(gtLabel, 1)
                else:
                    labelCount[gtLabel] += 1

        if not labelCount:
            return 0
                    
        # Find out the label with largest frequence  
        items = labelCount.items()
        items.sort(key = lambda labelCount:labelCount[1], reverse = True)
        key, value = items[0]
 
        labels = {
            LABEL_TABLECAPTION:     AMPLIFY_TABACCESSORY,
            LABEL_TABLEFOOTNOTE:    AMPLIFY_TABACCESSORY,
            LABEL_MATRIX:           AMPLIFY_MATRIX
        }
        strs = {
            LABEL_TABLECAPTION:     'amplified_tabaccessory',
            LABEL_TABLEFOOTNOTE:    'amplified_tabaccessory',
            LABEL_MATRIX:           'amplified_matrix'
        }

        ampLen = len(obPids) - len(frozenset(gtPids) & frozenset(obPids))
        pidRatio = float(ampLen) / self._charno
  
        if ((value > AMP_LABEL_RATIO * ampLen) & (key in labels)):
            amplify = Error_Label[labels[key]] * (1 / float(1 + exp(AMP_SIG_OFFSET - 10 * pidRatio)))
            self._labels['%s' %strs[key]] += 1
        else:     
            amplify = Error_Label[AMPLIFY_OTHER] * (1 / float(1 + exp(RDC_SIG_OFFSET - 10 * pidRatio)))
            self._labels['amplified_otherObj'] += 1    
        return amplify


    def _mapOb2Gt(self, pid): 
        lid = self._gtPage.getByPid(pid).getLid()
        ancestors = self._gtPage.lidToAncestors(lid)
        ancLabels = [ancestor.getLabel() for ancestor in ancestors]
   
        candiLabels = (LABEL_FORMULA, LABEL_MATRIX, LABEL_FIGURE, LABEL_LIST,       \
                       LABEL_TABLECAPTION, LABEL_TABLEFOOTNOTE, LABEL_TABLEBODY,    \
                       LABEL_PARAGRAPH)
        
        realLabel = LABEL_BODY
        for label in reversed(ancLabels):
            if label in candiLabels:
                realLabel = label
        return realLabel


        
        


   



            
        

        
    


    
    




